'''
Description: filter the unvaluable tRNA sequences of given .fa file
Note: set "-i 64" in general case.
'''
import optparse

usage = """python -m plastidUtilis.Filter -f <input.fasta> -i <minumum_length> -I <maximum_length> -o <output filename>
                                                                                                                    --Joe"""

parser = optparse.OptionParser(usage)
parser.add_option("-f", dest="input_file", help="fasta",
                  metavar="FILE", action="store", type="string")
parser.add_option("-i", dest="min", help="minimum length",
                  metavar="STRING", action="store", type="string")
parser.add_option("-I", dest="max", help="maximum length",
                  metavar="STRING", action="store", type="string")
parser.add_option("-o", dest="output", help="output file",
                  metavar="FILE", action="store", type="string")
(options, args) = parser.parse_args()

fa_dict = {}
fa_dict["dump"] = []
with open(options.input_file) as f:
    for line in f.readlines():
        if line.startswith(">"):
            header = line.strip()
            if header not in fa_dict.keys():
                fa_dict[header] = []
            elif header in fa_dict.keys():
                header = "dump"
        else:
            fa_dict[header].append(line.strip())
del fa_dict["dump"] 

with open(options.output, "w") as out:
    for k, v in fa_dict.items():
        fa_dict[k] = "".join(v)
        if len(fa_dict[k]) > int(options.min) and len(fa_dict[k]) < int(options.max):
            out.write(k + "\n")
            out.write(fa_dict[k] + "\n")
out.close()
                
        
        



'''this module is used to detect the abnormal CDS sequence annotated by Geseq
Note: input sequence should be in protein / amino acid format. 
'''
def AbnormalDetect(input, output):
    aa_dict = {}
    # aa_file = options.aa
    aa_file = input
    # out_file = options.output
    out_file = output
    with open(aa_file) as f:
        data = f.readlines()
        for line in data:
            if line.startswith(">"):
                aa_header = line.strip()
                aa_dict[aa_header] = []
            else:
                aa_dict[aa_header].append(line.strip())

    for k, v in aa_dict.items():
        aa_dict[k] = "".join(v)

    normal_dict = {}
    abnormal_dict = {}

    for k, v in aa_dict.items():
        # normal
        if v.endswith("*") and int(len(v.split("*"))) == 2:
            normal_dict[k.split()[0][1:]] = v
        else:
            # abnormal
            abnormal_dict[k.split()[0][1:]] = v
    with open(out_file, "w") as out:
        for k, v in abnormal_dict.items():
            if k in normal_dict.keys():
                # copy abnormal
                out.write(k+ "\t" + "copy_abnormal" + "\n")
            elif k not in normal_dict.keys():
                out.write(k + "\n")
    out.close()


def main():
    import optparse

    usage = """python -m plastidUtilis.AbnormalDetect -a <input_fasta> -o <output_filename>
                                                                                        --Joe"""
    parser = optparse.OptionParser(usage)
    parser.add_option("-a", dest="aa", help="input protein fasta",
                    metavar="FILE", action="store", type="string")
    parser.add_option("-o", dest="output", help="output filename",
                    metavar="OUT", action="store", type="string")
    (options, args) = parser.parse_args()

    AbnormalDetect(options.aa, options.output)


if __name__ == "__main__":
    main()



'''
Description: filter the unvaluable tRNA sequences of given .fa file
Note: "-d <delimiter>" parameter is used to retrieve the targeted ID instead of the whole header
'''
import optparse
usage = """python -m plastidUtilis.Longest -f <input.fasta> -d <delimiter> -o <output.fasta>
                                                                                    --Joe"""

parser = optparse.OptionParser(usage)
parser.add_option("-f", dest="input_fasta", help="input fasta",
                  metavar="FILE", action="store", type="string")
parser.add_option("-d", dest="delimiter", help="delimiter",
                  metavar="OUT", action="store", type="string")
parser.add_option("-o", dest="output", help="output fasta",
                  metavar="OUT", action="store", type="string")
(options, args) = parser.parse_args()

file_name = options.input_fasta
delimiter = str(options.delimiter)
output = options.output

fa_dict = {}
with open(file_name) as f:
    for line in f.readlines():
        if line.startswith(">"):
            seqname = line.strip()[1:]
            fa_dict[seqname] = []
        else:
            fa_dict[seqname].append(line.strip())
for k, v in fa_dict.items():
    fa_dict[k] = "".join(v)

keep_list = {}
for k in fa_dict.keys():
    gene = k.split(delimiter)[0]
    length = int(len(fa_dict[k]))
    if gene not in keep_list.keys():
        keep_list[gene] = k
    elif gene in keep_list and length > len(fa_dict[keep_list[gene]]):
        keep_list[gene] = k

with open(output, "w") as out:
    for v in keep_list.values():
        out.write(">" + v + "\n")
        out.write(fa_dict[v] + "\n")
out.close()


'''this module is parse the self-made TABLE file into GenBank format
Example TABLE: it should contain features like: 
1) nad1, nad2, nad5 gene: trans-splicing gene
#nad1
nad1_1	gene	361447	361833	-
nad1_1	CDS	361447	361833	-	1
nad1_1	note	trans_splicing	trans_splicing

2) tRNA gene
3) rRNA gene

!!! All the product should be filled before the GBK transformation
Dev: What I want to do is automatically filling!!!
'''
def TABLE2GBK(input, output):
    gene_dict = {}
    # with open(options.table) as f:
    with open(input, 'rt', encoding='utf-8') as f:   
        # if not set "utf-8", it will report 
        # UnicodeDecodeError: 'gbk' codec can't decode byte 0xaa in position 10: illegal multibyte sequence

        data = f.readlines()
        gene_num = 0
        for line in data:
            # not blank
            if line != "\n" and not line.startswith("#"):
                treat = line.split("\t")[1]
                if treat == 'gene':
                    gene_name = line.split("\t")[0] + "@" + str(gene_num)
                    gene_dict[gene_name] = {}
                    # start end orientation
                    gene_dict[gene_name]['gene'] = [line.split()[2], line.split()[3], line.split()[4]]
                    gene_num = gene_num + 1
                    continue
                elif treat == 'CDS' or treat == 'tRNA' or treat == 'rRNA':
                    if treat not in gene_dict[gene_name].keys():
                        gene_dict[gene_name][treat] = {}
                        exon = line.split()[5]
                        # start end orientation
                        gene_dict[gene_name][treat][exon] = [line.split()[2], line.split()[3], line.split()[4]]
                        continue
                    elif treat in gene_dict[gene_name].keys():
                        exon = line.split()[5]
                        # start end orientation
                        gene_dict[gene_name][treat][exon] = [line.split()[2], line.split()[3], line.split()[4]]
                        continue

                elif treat == 'note':
                    note = line.split("\t")[2]
                    print(note)
                    content = line.split("\t")[3].strip()
                    gene_dict[gene_name][note] = content
            

    # print gbk
    front_gap = " " * 5
    # with open(out_gbk, "w") as out:
    with open(output, "w") as out:
        for k, v in gene_dict.items():
            # gene
            rear_gap = " " * (21 - 5 - int(len("gene")))
            start = gene_dict[k]["gene"][0]
            end = gene_dict[k]["gene"][1]
            sign = gene_dict[k]["gene"][2]
            out.write(front_gap + "gene" + rear_gap)
            if sign == "+":
                out.write(start + ".." + end + "\n")
                out.write(21 * " " + '/gene="%s"' % k.split("@")[0] + "\n")
            elif sign == "-":
                out.write("complement(" + start + ".." + end + ")" + "\n")
                out.write(21 * " " + '/gene="%s"' % k.split("@")[0] + "\n")

            # CDS
            if "CDS" in gene_dict[k].keys() and "trans_splicing" not in gene_dict[k].keys():
                rear_gap = " " * (21 - 5 - int(len("CDS")))
                out.write(front_gap + "CDS" + rear_gap)
                # no intron
                exon_num = int(len(gene_dict[k]["CDS"].keys()))
                if exon_num == 1:
                    if sign == "+":
                        out.write(start + ".." + end + "\n")
                        out.write(21 * " " + '/gene="%s"' % k.split("@")[0] + "\n")
                    elif sign == "-":
                        out.write("complement(" + start + ".." + end + ")" + "\n")
                        out.write(21 * " " + '/gene="%s"' % k.split("@")[0] + "\n")
                # intron
                elif exon_num > 1:
                    exon_list = []
                    if sign == "+":
                        for i in range(1, exon_num + 1, 1):
                            start = gene_dict[k]["CDS"][str(i)][0]
                            end = gene_dict[k]["CDS"][str(i)][1]
                            exon_list.append(start + ".." + end)
                        out.write("join(" + ",".join(exon_list) + ")" + "\n")
                    elif sign == "-":
                        for i in range(exon_num, 0, -1):
                            start = gene_dict[k]["CDS"][str(i)][0]
                            end = gene_dict[k]["CDS"][str(i)][1]
                            exon_list.append(start + ".." + end)
                        out.write("complement(join(" + ",".join(exon_list) + "))" + "\n")
                    out.write(21 * " " + '/gene="%s"' % k.split("@")[0] + "\n")

                # product
                if "product" in gene_dict[k].keys():
                    out.write(21 * " " + '/product="%s"' % gene_dict[k]["product"] + "\n")

                # translation
                if "translation" in gene_dict[k].keys():
                    translation = '/translation="%s"' % gene_dict[k]["translation"]
                    translation_list = [translation[i: i + 58] for i in range(0, len(translation), 58)]
                    for i in range(0, len(translation_list)):
                        out.write(21 * " " + translation_list[i] + "\n")

                # exon and intron
                exon_rear_gap = " " * (21 - 5 - int(len("exon")))
                intron_rear_gap = " " * (21 - 5 - int(len("intron")))
                if exon_num > 1:
                    if sign == "+":
                        for i in range(1, exon_num + 1, 1):
                            start = gene_dict[k]["CDS"][str(i)][0]
                            end = gene_dict[k]["CDS"][str(i)][1]
                            out.write(front_gap + "exon" + exon_rear_gap + start + ".." + end + "\n")
                            out.write(21 * " " + '/gene="%s"' % k.split("@")[0] + "\n")
                            out.write(21 * " " + '/number="%s"' % str(i) + "\n")
                            intron_num = i
                            if intron_num < exon_num:
                                intron_start = int(end) + 1
                                intron_end = int(gene_dict[k]["CDS"][str(i + 1)][0]) - 1
                                out.write(
                                    front_gap + "intron" + intron_rear_gap + str(intron_start) + ".." + str(
                                        intron_end) + "\n")
                                out.write(21 * " " + '/gene="%s"' % k.split("@")[0] + "\n")
                                out.write(21 * " " + '/number="%s"' % str(intron_num) + "\n")

                    if sign == "-":
                        for i in range(exon_num, 0, -1):
                            start = gene_dict[k]["CDS"][str(i)][0]
                            end = gene_dict[k]["CDS"][str(i)][1]
                            out.write(front_gap + "exon" + exon_rear_gap + "complement(" + start + ".." + end + ")" + "\n")
                            out.write(21 * " " + '/gene="%s"' % k.split("@")[0] + "\n")
                            out.write(21 * " " + '/number="%s"' % str(i) + "\n")
                            intron_num = i - 1
                            if intron_num > 0:
                                intron_start = int(end) + 1
                                intron_end = int(gene_dict[k]["CDS"][str(i - 1)][0]) - 1
                                out.write(
                                    front_gap + "intron" + intron_rear_gap + "complement(" + str(intron_start) + ".." + str(
                                        intron_end) + ")" + "\n")
                                out.write(21 * " " + '/gene="%s"' % k.split("@")[0] + "\n")
                                out.write(21 * " " + '/number="%s"' % str(intron_num) + "\n")

            # trans_splicing
            elif "CDS" in gene_dict[k].keys() and "trans_splicing" in gene_dict[k].keys():
                exon_rear_gap = " " * (21 - 5 - int(len("exon")))
                intron_rear_gap = " " * (21 - 5 - int(len("intron")))
                exon_num = int(len(gene_dict[k]["CDS"].keys()))
                # one exon
                if exon_num == 1:
                    out.write(front_gap + "exon" + exon_rear_gap)
                    if sign == "+":
                        out.write(start + ".." + end + "\n")
                    elif sign == "-":
                        out.write("complement(" + start + ".." + end + ")" + "\n")
                    out.write(21 * " " + '/gene="%s"' % k.split("@")[0] + "\n")
                    out.write(21 * " " + '/number="%s"' % str(list(gene_dict[k]["CDS"].keys())[0]) + "\n")

                # exon and intron
                if exon_num > 1:
                    intron_num = 1
                    if sign == "+":
                        num_list = list(sorted(list([gene_dict[k]["CDS"].keys()])[0]))
                        for i in num_list:
                            start = gene_dict[k]["CDS"][str(i)][0]
                            end = gene_dict[k]["CDS"][str(i)][1]
                            out.write(front_gap + "exon" + exon_rear_gap + start + ".." + end + "\n")
                            out.write(21 * " " + '/gene="%s"' % k.split("@")[0] + "\n")
                            out.write(21 * " " + '/number="%s"' % str(i) + "\n")
                            if intron_num < exon_num:
                                intron_start = int(end) + 1
                                intron_end = int(gene_dict[k]["CDS"][str(int(i) + 1)][0]) - 1
                                out.write(
                                    front_gap + "intron" + intron_rear_gap + str(intron_start) + ".." + str(
                                        intron_end) + "\n")
                                out.write(21 * " " + '/gene="%s"' % k.split("@")[0] + "\n")
                                out.write(21 * " " + '/number="%s"' % str(intron_num) + "\n")
                            intron_num = intron_num + 1

                    if sign == "-":
                        num_list = list(sorted(list([gene_dict[k]["CDS"].keys()])[0], reverse=True))
                        for i in num_list:
                            start = gene_dict[k]["CDS"][str(i)][0]
                            end = gene_dict[k]["CDS"][str(i)][1]
                            out.write(front_gap + "exon" + exon_rear_gap + "complement(" + start + ".." + end + ")" + "\n")
                            out.write(21 * " " + '/gene="%s"' % k.split("@")[0] + "\n")
                            out.write(21 * " " + '/number="%s"' % str(i) + "\n")
                            if intron_num < exon_num:
                                intron_start = int(end) + 1
                                intron_end = int(gene_dict[k]["CDS"][str(int(i) - 1)][0]) - 1
                                out.write(
                                    front_gap + "intron" + intron_rear_gap + "complement(" + str(intron_start) + ".." + str(
                                        intron_end) + ")" + "\n")
                                out.write(21 * " " + '/gene="%s"' % k.split("@")[0] + "\n")
                                out.write(21 * " " + '/number="%s"' % str(intron_num) + "\n")
                            intron_num = intron_num + 1


            # tRNA
            elif "tRNA" in gene_dict[k].keys():
                tRNA_rear_gap = " " * (21 - 5 - int(len("tRNA")))
                out.write(front_gap + "tRNA" + tRNA_rear_gap)
                # no intron
                exon_num = int(len(gene_dict[k]["tRNA"].keys()))
                if exon_num == 1:
                    if sign == "+":
                        out.write(start + ".." + end + "\n")
                        out.write(21 * " " + '/gene="%s"' % k.split("@")[0] + "\n")
                    elif sign == "-":
                        out.write("complement(" + start + ".." + end + ")" + "\n")
                        out.write(21 * " " + '/gene="%s"' % k.split("@")[0] + "\n")
                # intron
                elif exon_num > 1:
                    exon_list = []
                    if sign == "+":
                        for i in range(1, exon_num + 1, 1):
                            start = gene_dict[k]["tRNA"][str(i)][0]
                            end = gene_dict[k]["tRNA"][str(i)][1]
                            exon_list.append(start + ".." + end)
                        out.write("join(" + ",".join(exon_list) + ")" + "\n")
                    elif sign == "-":
                        for i in range(exon_num, 0, -1):
                            start = gene_dict[k]["tRNA"][str(i)][0]
                            end = gene_dict[k]["tRNA"][str(i)][1]
                            exon_list.append(start + ".." + end)
                        out.write("complement(join(" + ",".join(exon_list) + "))" + "\n")
                    out.write(21 * " " + '/gene="%s"' % k.split("@")[0] + "\n")
                    
                # product
                if "product" in gene_dict[k].keys():
                    out.write(21 * " " + '/product="%s"' % gene_dict[k]["product"] + "\n")

                # exon and intron
                exon_rear_gap = " " * (21 - 5 - int(len("exon")))
                intron_rear_gap = " " * (21 - 5 - int(len("intron")))
                if exon_num > 1:
                    if sign == "+":
                        for i in range(1, exon_num + 1, 1):
                            start = gene_dict[k]["tRNA"][str(i)][0]
                            end = gene_dict[k]["tRNA"][str(i)][1]
                            out.write(front_gap + "exon" + exon_rear_gap + start + ".." + end + "\n")
                            out.write(21 * " " + '/gene="%s"' % k.split("@")[0] + "\n")
                            out.write(21 * " " + '/number="%s"' % str(i) + "\n")
                            intron_num = i
                            if intron_num < exon_num:
                                intron_start = int(end) + 1
                                intron_end = int(gene_dict[k]["tRNA"][str(i + 1)][0]) - 1
                                out.write(
                                    front_gap + "intron" + intron_rear_gap + str(intron_start) + ".." + str(
                                        intron_end) + "\n")
                                out.write(21 * " " + '/gene="%s"' % k.split("@")[0] + "\n")
                                out.write(21 * " " + '/number="%s"' % str(intron_num) + "\n")

                    if sign == "-":
                        for i in range(exon_num, 0, -1):
                            start = gene_dict[k]["tRNA"][str(i)][0]
                            end = gene_dict[k]["tRNA"][str(i)][1]
                            out.write(front_gap + "exon" + exon_rear_gap + "complement(" + start + ".." + end + ")" + "\n")
                            out.write(21 * " " + '/gene="%s"' % k.split("@")[0] + "\n")
                            out.write(21 * " " + '/number="%s"' % str(i) + "\n")
                            intron_num = i - 1
                            if intron_num > 0:
                                intron_start = int(end) + 1
                                intron_end = int(gene_dict[k]["tRNA"][str(i - 1)][0]) - 1
                                out.write(
                                    front_gap + "intron" + intron_rear_gap + "complement(" + str(intron_start) + ".." + str(
                                        intron_end) + ")" + "\n")
                                out.write(21 * " " + '/gene="%s"' % k.split("@")[0] + "\n")
                                out.write(21 * " " + '/number="%s"' % str(intron_num) + "\n")


            # rRNA
            elif "rRNA" in gene_dict[k].keys():
                rRNA_rear_gap = " " * (21 - 5 - int(len("rRNA")))
                out.write(front_gap + "rRNA" + rRNA_rear_gap)
                if sign == "+":
                    out.write(start + ".." + end + "\n")
                    out.write(21 * " " + '/gene="%s"' % k.split("@")[0] + "\n")
                elif sign == "-":
                    out.write("complement(" + start + ".." + end + ")" + "\n")
                # product
                if "product" in gene_dict[k].keys():
                    out.write(21 * " " + '/product="%s"' % gene_dict[k]["product"] + "\n")
                    
            out.write("\n")
    out.close()


def main():
    import optparse
    usage = """python -m plastidUtilis.Table2GBK -t <self-made table file> -o <.gb> 
                                                                      --Joe"""

    parser = optparse.OptionParser(usage)
    parser.add_option("-t", dest="table", help="input annotation table",
                    metavar="FILE", action="store", type="string")
    parser.add_option("-o", dest="outname", help="output filename",
                    metavar="OUT", action="store", type="string")
    (options, args) = parser.parse_args()

    TABLE2GBK(options.table, options.outname)

if __name__ == "__main__":
    main()


'''this module is used to append up/down-stream sequence to the target sequence
version1: retrive extending sequences based on BED file generated by JCVI
JCVI:
python -m jcvi.formats.gff bed --type=gene --key=gene GeSeqJob.gff3 -o GeSeq.bed
Note: 
1) JCVI will transform 1-based GFF3 to 0-based BED, which is also suitable for TBtools
2) due to trans-splicing, it might report the unusual range like:
"12588   307844" which indicates that the total gene length is 295256.
'''

from Bio import SeqIO
from Bio.SeqRecord import SeqRecord
from Bio.Seq import Seq


def ParsePlastid(filename):
    '''this file is used to parse the plastid genome,
    Note: only consider single-loop plastid'''
    seq_id = ''
    sequence = ''
    with open(filename, 'rt') as input:
        for record in SeqIO.parse(input, 'fasta'):
            seq_id = record.id
            sequence = str(record.seq)
    return seq_id, sequence


def RetrieveAbnormal(filename, sequence, N, outputname):
    '''Read BED file of abnormal gene and retrieve the expanded range of sequence,
    "expanded" means that up- and down-stream with N bp
    Note: Because this module apply "SeqIO.write" to conduct the saving process,
    it might not be smooth to paste the target sequence to tools like sublime text to further down-stream analysis
    '''
    abnormals = {}
    with open(filename, 'rt') as input:
        for line in input: # "fragment", "copy-abnormal" filter were not set
            ab = line.strip('\n').split('\t')
            start_pos = int(ab[1]) - int(N)
            end_pos = int(ab[2]) + int(N)
            
            if ab[3] in ['nad1', 'nad2', 'nad5'] or "fragment" in ab[3] or "copy_abnormal" in ab[3]: # ignore
                continue

            if ab[-1] == '+':    
                # if seq is on the "+" strand, its down-stream will be appended
                # "-" strand, its up-stream will be appended
                abnormals[ f'{ab[3]}\t{ab[1]}:{ab[1]}\t{ab[2]}:{end_pos}\t{ab[-1]}' ] = sequence[int(ab[1]):end_pos]  # {header}:sequence
            else:
                abnormals[ f'{ab[3]}\t{ab[1]}:{start_pos}\t{ab[2]}:{ab[2]}\t{ab[-1]}' ] = Seq(sequence[start_pos:int(ab[2])]).reverse_complement()
    Records = []
    for record in list(abnormals.items()):
        tmpRecord = SeqRecord(Seq(record[1]), id=record[0], description='')
        Records.append(tmpRecord)
    SeqIO.write(Records, outputname, 'fasta')


def main():
    import optparse

    usage = """python -m plastidUtilis.SequenceAppend -f <input_fasta> -i <input_BED> -o <output_filename>"""
    parser = optparse.OptionParser(usage)
    parser.add_option("-f", dest="fasta", help="input plastid/mitochondia genome fasta",
                    metavar="FILE", action="store", type="string")
    parser.add_option("-i", dest="bed", help="input BED of abnormal gene/fragment",
                    metavar="FILE", action="store", type="string")
    parser.add_option("-n", "--number", help="the number of flanking bp",
                    metavar="NUMBER", action="store", type="int")
    parser.add_option("-o", dest="output", help="output fasta name",
                    metavar="OUT", action="store", type="string")
    (options, args) = parser.parse_args()

    _, sequence = ParsePlastid(options.fasta)
    RetrieveAbnormal(options.bed, sequence, options.number, options.output)


if __name__ == "__main__":
    main()
